//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Model/JobModel.cpp
//! @brief     Implements class JobModel
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Model/JobModel.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/PointwiseAxisItem.h"
#include "GUI/Model/Job/FitSuiteItem.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Model/JobQueueData.h"
#include "GUI/Model/Model/ParameterTreeUtils.h"
#include "GUI/Model/Model/SessionXML.h"
#include "GUI/Model/Sample/ItemWithMaterial.h"
#include "GUI/Model/Sample/MultiLayerItem.h"

#include "GUI/Support/Data/SessionFlags.h"
#include "GUI/Support/XML/UtilXML.h"

#include "GUI/Util/Error.h"

namespace {
namespace Tags {

const QString JobItemTag("JobItem");

} // namespace Tags
} // namespace

JobModel::JobModel(QObject* parent)
    : SessionModel(GUI::Session::XML::JobModelTag, parent)
    , m_queue_data(nullptr)
{
    m_queue_data = new JobQueueData(this);
    connect(m_queue_data, &JobQueueData::focusRequest, this, &JobModel::focusRequest);
    connect(m_queue_data, &JobQueueData::globalProgress, this, &JobModel::globalProgress);
    setObjectName(GUI::Session::XML::JobModelTag);
}

JobModel::~JobModel()
{    
    delete m_queue_data;

#ifdef NEW_JOB_MODEL
    // jobs should be deleted explicitly, because they are not in tree anymore
    qInfo() << "destruction of JobModel";
    for(auto* jobItem : m_jobItems)
        delete jobItem;
#endif
}

JobItem* JobModel::getJobItemForIdentifier(const QString& identifier)
{
    for (auto* jobItem : jobItems())
        if (jobItem->getIdentifier() == identifier)
            return jobItem;

    return nullptr;
}

#ifdef NEW_JOB_MODEL
JobItem *JobModel::createJobItem()
{
    JobItem* jobItem = new JobItem();
    jobItem->setParentAndModel(nullptr, this);
    m_jobItems.push_back(jobItem);
    return jobItem;
}
#endif

//! Main method to add a job
JobItem* JobModel::addJob(const MultiLayerItem* sampleItem, const InstrumentItem* instrumentItem,
                          const RealItem* realItem, const SimulationOptionsItem& optionItem)
{
    ASSERT(sampleItem);
    ASSERT(instrumentItem);

#ifdef NEW_JOB_MODEL
    JobItem* jobItem = createJobItem();
#else
    JobItem* jobItem = insertItem<JobItem>();
#endif
    jobItem->setJobName(generateJobName());
    jobItem->setIdentifier(QUuid::createUuid().toString());

    jobItem->copySampleIntoJob(sampleItem);
    jobItem->copyInstrumentIntoJob(instrumentItem);
    ParameterTreeBuilder(jobItem, true).build();
    jobItem->copySimulationOptionsIntoJob(optionItem);
    jobItem->createSumulatedDataItem();

    if (realItem) {
        jobItem->copyRealItemIntoJob(realItem);
        jobItem->adjustReaDataToJobInstrument();

        jobItem->createDiffDataItem();
        jobItem->createFitContainers();
    }

    emit jobAdded();
    return jobItem;
}

QVector<JobItem*> JobModel::jobItems() const
{
#ifdef NEW_JOB_MODEL
    return m_jobItems;
#else
    return topItems<JobItem>();
#endif
}

//! restore instrument and sample model from backup for given JobItem
void JobModel::restore(JobItem* jobItem)
{
    jobItem->parameterContainerItem()->restoreBackupValues();
}

bool JobModel::hasUnfinishedJobs()
{
    if (m_queue_data->hasUnfinishedJobs())
        return true;

    for (auto* jobItem : jobItems())
        if (jobItem->getStatus() == JobStatus::Fitting)
            return true;

    return false;
}

// used before reading from xml file
void JobModel::clear()
{
    for (auto* job : jobItems())
        removeJob(job);

    SessionModel::clear();
}

QVector<DataItem*> JobModel::dataItems() const
{
    QVector<DataItem*> result;

    for (auto* jobItem : jobItems()) {
        if (auto* dataItem = jobItem->simulatedDataItem())
            result.push_back(dataItem);

        if (auto* real_data = dynamic_cast<RealItem*>(jobItem->realItem())) {
            if (auto* data_item = real_data->dataItem())
                result.push_back(data_item);
            if (auto* native_data = real_data->nativeData())
                result.push_back(native_data);
        }
    }

    return result;
}

void JobModel::writeTo(QXmlStreamWriter *writer)
{
#ifdef NEW_JOB_MODEL

    ASSERT(writer);
    writer->writeStartElement(GUI::Session::XML::JobModelTag);

    for (auto* job : jobItems()) {
        writer->writeStartElement(GUI::Session::XML::ItemTag);
        writer->writeAttribute(GUI::Session::XML::ModelTypeAttribute, Tags::JobItemTag);

        writer->writeAttribute(GUI::Session::XML::Tags::TagAttribute, "rootTag");
        writer->writeAttribute(GUI::Session::XML::Tags::DisplayNameAttribute,
                               job->roleProperty(SessionFlags::DisplayNameRole).toString());

        GUI::Session::XML::writeItemAndChildItems(writer, job);

        writer->writeEndElement(); // ItemTag
    }
    writer->writeEndElement(); // JobModelTag

#else
    SessionModel::writeTo(writer);
#endif
}

void JobModel::readFrom(QXmlStreamReader* reader, MessageService* messageService /*= 0*/)
{
#ifdef NEW_JOB_MODEL
    // explict realization, following SessionModel::readFrom
    ASSERT(reader);

    // we should check top-level tag
    if (reader->name() != QString(GUI::Session::XML::JobModelTag))
        throw Error("JobModel::readFrom() -> Format error in p1");

    beginResetModel();
    clear();

//    GUI::Session::XML::readItems(reader, m_root_item, QString(), messageService);
    while (reader->readNextStartElement()) {

        if (reader->name().toString() == GUI::Session::XML::ItemTag) {
            const QString model_type =
                    reader->attributes().value(GUI::Session::XML::Tags::ModelTypeAttribute).toString();

            if (model_type == Tags::JobItemTag) {

                JobItem* jobItem = createJobItem();

                reader->readNextStartElement();
                if (reader->name().toString() == GUI::Session::XML::Tags::NonSessionItemData) {
                    jobItem->readNonSessionItems(reader);
                    GUI::Session::XML::gotoEndElementOfTag(reader, GUI::Session::XML::Tags::NonSessionItemData);
                }
            }
            GUI::Session::XML::gotoEndElementOfTag(reader, GUI::Session::XML::ItemTag);
        }
    }

    if (reader->hasError())
        throw Error(reader->errorString());
    endResetModel();
#else
    SessionModel::readFrom(reader, messageService);
#endif

    // Set & check instrument links. JobItem and realItem have to reference the same instrument
    // (which is the one contained in this job item, not contained in the instrument model)
    for (auto* jobItem : jobItems()) {
        if (RealItem* refItem = jobItem->realItem())
            refItem->linkToInstrument(jobItem->instrumentItem());

        // Create the not stored parameter tuning tree
        ParameterTreeBuilder(jobItem, false).build();
    }
}

void JobModel::runJob(JobItem* jobItem)
{
    m_queue_data->runJob(jobItem);
}

void JobModel::cancelJob(JobItem* jobItem)
{
    m_queue_data->cancelJob(jobItem->getIdentifier());
}

void JobModel::removeJob(JobItem* jobItem)
{
    ASSERT(jobItem);
    m_queue_data->removeJob(jobItem->getIdentifier());

#ifdef NEW_JOB_MODEL
    delete jobItem;
    m_jobItems.removeOne(jobItem);
#else
    removeRows(jobItem->index().row(), 1, QModelIndex());
#endif
}

QString JobModel::generateJobName()
{
    int maxJobIndex = 0;
    for (const auto* jobItem : jobItems()) {
        if (jobItem->jobName().startsWith("job")) {
            maxJobIndex = std::max(maxJobIndex, jobItem->jobName().mid(3).toInt());
        }
    }
    return QString("job%1").arg(maxJobIndex + 1);
}
