//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/ba3d/def.cpp
//! @brief     Definitions
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/ba3d/def.h"

namespace GUI::RealSpace {

F3::F3()
    : F3(0, 0, 0)
{
}

F3::F3(float v)
    : F3(v, v, v)
{
}

F3::F3(float x_, float y_, float z_)
    : x(x_)
    , y(y_)
    , z(z_)
{
}

F3::F3(QVector3D const& v)
    : F3(v.x(), v.y(), v.z())
{
}

float F3::length() const
{
    return QVector3D(*this).length();
}

F3 F3::normalized() const
{
    return QVector3D(*this).normalized();
}

F3 F3::interpolateTo(const F3& to, float rat) const
{
    return *this * (1 - rat) + to * rat;
}

GUI::RealSpace::F3::operator QVector3D() const
{
    return QVector3D(x, y, z);
}

F3 const F3::_0(0, 0, 0), F3::_1(1, 1, 1), F3::_x(1, 0, 0), F3::_y(0, 1, 0), F3::_z(0, 0, 1);

F3 cross(const F3& v1, const F3& v2)
{
    return QVector3D::crossProduct(v1, v2);
}

float dot(const F3& v1, const F3& v2)
{
    return QVector3D::dotProduct(v1, v2);
}

F3 operator+(const F3& v)
{
    return v;
}

F3 operator-(const F3& v)
{
    return F3::_0 - v;
}

F3 operator*(const F3& v, float f)
{
    return F3(v.x * f, v.y * f, v.z * f);
}

F3 operator+(const F3& _1, const F3& _2)
{
    return F3(_1.x + _2.x, _1.y + _2.y, _1.z + _2.z);
}

F3 operator-(const F3& _1, const F3& _2)
{
    return F3(_1.x - _2.x, _1.y - _2.y, _1.z - _2.z);
}


Range::Range(float r1, float r2)
    : min(qMin(r1, r2))
    , max(qMax(r1, r2))
{
}

float Range::size() const
{
    return max - min;
}

float Range::mid() const
{
    return (min + max) / 2;
}


VectorRange::VectorRange(Range x_, Range y_, Range z_)
    : x(x_)
    , y(y_)
    , z(z_)
{
}

VectorRange::VectorRange(F3 _1, F3 _2)
    : x(Range(_1.x, _2.x))
    , y(Range(_1.y, _2.y))
    , z(Range(_1.z, _2.z))
{
}

F3 VectorRange::size() const
{
    return F3(x.size(), y.size(), z.size());
}

F3 VectorRange::mid() const
{
    return F3(x.mid(), y.mid(), z.mid());
}

float VectorRange::length() const
{
    return F3(x.size(), y.size(), z.size()).length();
}

} // namespace GUI::RealSpace
